/* The software provided here is released by the National
 * Institute of Standards and Technology (NIST), an agency of
 * the U.S. Department of Commerce, Gaithersburg MD 20899,
 * USA.  The software bears no warranty, either expressed or
 * implied. NIST does not assume legal liability nor
 * responsibility for a User's use of the software or the
 * results of such use.
 * Please note that within the United States, copyright
 * protection, under Section 105 of the United States Code,
 * Title 17, is not available for any work of the United
 * States Government and/or for any works created by United
 * States Government employees. User acknowledges that this
 * software contains work which was created by NIST employees
 * and is therefore in the public domain and not subject to
 * copyright.  The User may use, distribute, or incorporate
 * this software provided the User acknowledges this via an
 * explicit acknowledgment of NIST-related contributions to
 * the User's work. User also agrees to acknowledge, via an
 * explicit acknowledgment, that any modifications or
 * alterations have been made to this software before
 * redistribution.
 * --------------------------------------------------------------------
 *
 * Change History:
 * Simson L. Garfinkel - May 1, 2008
 *   - Major rewrite using new function
 * Simson L. Garfinkel - simsong@acm.org - August 21, 2006
 *   - Re-implemented based on Douglas White's original Perl code.
 *   - First 4096-bytes of file describes parameters
 *   - Single executable for both 128-bit and 160-bit Bloom filters.
 *   - Uses memmap() for handling 512MB bloom filters. 
 * Douglas White - douglas.white@nist.gov - June 21, 2003
 *   Original implementation in perl.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <errno.h>
#include <err.h>
#include <ctype.h>

#include "bloom.h"

#define BLOOM_VECTOR_OFFSET 4096	/* size of one page */

static int *hexcharvals = 0;
int nsrl_klookup_counters[32];	/* keep track of the number of times 0, 1, 2 functions used for lookup */

static void nsrl_bloom_init()
{
    if(hexcharvals==0){
	/* Need to initialize this */
	int i;
	hexcharvals = calloc(sizeof(int),256);
	for(i=0;i<10;i++){
	    hexcharvals['0'+i] = i;
	}
	for(i=10;i<16;i++){
	    hexcharvals['A'+i-10] = i;
	    hexcharvals['a'+i-10] = i;
	}
	memset(nsrl_klookup_counters,0,sizeof(nsrl_klookup_counters));
    }
}


/* Convert a hex representation to binary, and return
 * the number of bits converted.
 */
int nsrl_hex2bin(unsigned char *binbuf,const char *hex)
{
    if(nsrl_debug)printf("in nsrl_hex2bin with %s\n", hex); 
    int bits = 0;
    if(hexcharvals==0) nsrl_bloom_init();
    while(hex[0] && hex[1]){
	*binbuf++ = ((hexcharvals[(unsigned char)hex[0]]<<4) |
		     hexcharvals[(unsigned char)hex[1]]);
	hex  += 2;
	bits += 8;
    }
    return bits;
}

const char *nsrl_hexbuf(char *dst,int dst_len,const unsigned char *bin,int bytes,int flag)
{
    int charcount = 0;
    const char *start = dst;		// remember where the start of the string is
    const char *fmt = (flag & NSRL_HEXBUF_UPPERCASE) ? "%02X" : "%02x";

    if(hexcharvals==0) nsrl_bloom_init();
    *dst = 0;				// begin with null termination
    while(bytes>0 && dst_len > 3){
	int add_spaces = 0;

	sprintf(dst,fmt,*bin); // convert the next byte
	dst += 2;
	bin += 1;
	dst_len -= 2;
	bytes--;
	charcount++;			// how many characters
	
	if(flag & NSRL_HEXBUF_SPACE2) add_spaces = 0;
	if((flag & NSRL_HEXBUF_SPACE4) && charcount%2==0){
	    *dst++ = ' ';
	    *dst   = '\000';
	    dst_len -= 1;
	}
    }
    return start;			// return the start
}


int nsrl_debug = 0;

static inline void set_bloom_bit(nsrl_bloom *b,unsigned int bit)
{
    int q = 1<<(bit%8);
    b->vector[bit/8] |= q;
    if(nsrl_debug>1) printf("bit/8=%d  adding %d\n",bit/8,q);
}

static inline int get_bloom_bit(const nsrl_bloom *b,unsigned int bit)
{
    int q = 1<<(bit%8);
    //if(nsrl_debug) printf("bit/8=%d  checking %d\n",bit/8,q);
    return (b->vector[bit/8] & q) ? 1 : 0;
}


void nsrl_bloom_info(char *buf,int buflen, const nsrl_bloom *b)
{
    snprintf(buf,buflen,
	     "nsrl_bf_version:2\n"
	     "hash_bits:%d\n"
	     "bloom_bits:%d\n"
	     "k:%d\n"
	     "added_items:%d\n"
	     "comment:%s\n",
	     b->hash_bits, b->bloom_bits,b->k,b->added_items, b->comment);
}

void nsrl_calc_histogram(const nsrl_bloom *b,unsigned int counts[256])
{
    int i;
    unsigned int byteswbits = 0;
	memset(counts,0,sizeof(counts));
    for(i=0;i<b->vector_bytes;i++){
    	if(b->vector[i]) {
			counts[b->vector[i]]++;
			byteswbits++;}
	}
}

void nsrl_print_histogram(const nsrl_bloom *b,const unsigned int counts[256])
{
	unsigned int totalbits = 0;
    unsigned int bitcount = 0;
    int i,j,mask;
    float top, bottom, percent = 0.0;
    for(i=1;i<256;i++){
	if(counts[i] > 0){
	    for(j = 0; j<8; j++){
		mask = (1 << j);
		if(i&mask){ bitcount += counts[i]; } 
	    }
	}
    }
    if(b->bloom_bits < 32) totalbits = 1<<b->bloom_bits;
    else{ totalbits--;}
    printf("%u bits total, %u bits set\n",totalbits, bitcount);
    top = (float)bitcount;
    bottom = (float)totalbits;
    percent = top/bottom;
    percent *= 100;
    printf("%f percent full\n", percent);
}


static int nsrl_bloom_info_update(nsrl_bloom *b)
{
    if(b->fd==0) return 0;		/* no open file to update */
    char buf[BLOOM_VECTOR_OFFSET];
    memset(buf,0,sizeof(buf));
    nsrl_bloom_info(buf,sizeof(buf),b);
    if (lseek(b->fd,0,SEEK_SET)) return -1;
    if (write(b->fd,buf,BLOOM_VECTOR_OFFSET)!=BLOOM_VECTOR_OFFSET) return -1;
    return 0;
}

/* Return the ith bit of a blob */
static inline unsigned int get_bit(const unsigned char *buf,unsigned int bit)
{
    return (buf[bit/8] & (1<<bit%8)) ? 1 : 0;
}


void nsrl_bloom_add(nsrl_bloom *b,const unsigned char *hash)
{
    /* Apply each round of nist_function128 and set the bits in the vector */
    unsigned int i,j;
    if (nsrl_debug>1) { printf("VECTORS:"); }
    for(i=0;i<b->k;i++){
	unsigned int offset = i * b->bloom_bits;
	unsigned int v = 0;
	for(j=0;j<b->bloom_bits;j++){
	    v = (v<<1) | get_bit(hash,offset+j);
	}
	if(nsrl_debug>1){printf("Setting bit: %u.\n",v);}
	set_bloom_bit(b,v);
	if(nsrl_debug>1) { printf("\t%08x",v);if(i%3==2) putchar('\n');}
    }
    b->added_items++;
    if(b->added_items % 1000==0) nsrl_bloom_info_update(b);
    if (nsrl_debug>1) { printf("\n"); }
}

void walkfile(nsrl_bloom *b, const char *fname){
    FILE *walk_file=0;
    char linebuf[256];
    char shatoquery[41];
    unsigned char hashbuf[120];
    unsigned int	count = 0, found =0;
    shatoquery[40] = '\0';
    walk_file = fopen(fname, "r");
    if(!walk_file){
	perror(fname);
	exit(1);
    }
	
    while(!feof(walk_file) && fgets(linebuf, sizeof(linebuf), walk_file)){
	if(!strncmp(linebuf, "SHA1: ", 6)){//valid sha1 line
	    strncpy(shatoquery, &linebuf[6], 40);
	    nsrl_hex2bin(hashbuf, shatoquery); //convert to binary for query	
	    count++;
	    if(nsrl_bloom_query(b, hashbuf)){
		//			printf("%s present\n", shatoquery);
		found++;
	    }
	    //		else{ printf("%s absent\n", shatoquery);}
	} //end if loop for valid sha1
    }//end while !feof loop
    fclose(walk_file);
    printf("%u found, %u searched.  %f %%coverage.\n", found, count, (((float)found/(float)count)*100));
}

void queryfile(nsrl_bloom *b,const char *fname)
{
    char linebuf[256];
    int  nsrl_file = 1;
    FILE *load_file = 0;

    int load_line = 0;
    if(strcmp(fname,"-")==0){
	load_file = stdin;
    }
    else {
	load_file = fopen(fname,"r");
    }
    if(!load_file){
	perror(fname);
	exit(1);
    }
    while(!feof(load_file) && fgets(linebuf,sizeof(linebuf),load_file)){
	unsigned char hashbuf[120];
	int found_bits;
	char *hash_start=0;
	load_line ++;
	linebuf[sizeof(linebuf)-1] = 0;	/* good hygine */
	nsrl_file = 1;//copied function from nsrl_load
	if(nsrl_file){
	    if(linebuf[0]=='"'){hash_start=&linebuf[1];}
	    else{hash_start=linebuf;} 
	    if(b->hash_bits==128){
		hash_start[32]='\000';	/* terminate */
	    }
	    else {
		hash_start[40]='\000';
	    }
	}

	if(!isxdigit(*hash_start)){
	    fprintf(stderr,"invalid hash: %s\n",hash_start);
	    continue; /* line doesn't begin with hash */
	}
	found_bits = nsrl_hex2bin(hashbuf,hash_start);
	if(found_bits != b->hash_bits){
	    char *c = index(linebuf,'\n');
	    if(c) *c = '\000';
	    fprintf(stderr,"Input format error.\n");
	    fprintf(stderr,"%s line %d: '%s' has %d bits (%d expected)\n",
		    fname,load_line,linebuf,found_bits,b->hash_bits);
	    exit(1);
	}
	if(nsrl_bloom_query(b,hashbuf)){
	    printf("%s present\n",hash_start);
	}
	else {
	    printf("%s absent\n",hash_start);
	}
    }
#ifdef SIGINFO
    signal(SIGINFO,0);
#endif
}
/* nsrl_bloom_query:
 * Check each round of nist_function128 in the vector. If any are not set,
 * then the hash is not in the bloom filter.
 *
 * Returns: 1 if present, 0 if not present
 */
int nsrl_bloom_query(nsrl_bloom *b,const unsigned char *hash)
{
    int i,j;
    if (nsrl_debug>1) { printf("VECTORS:"); }

    for(i=0;i<b->k;i++){		/* i is which vector function to query */
	nsrl_klookup_counters[i]++;		/* increment the  */

	int offset = i * b->bloom_bits;
	unsigned int v = 0;		/* v is the bit in the vector that is to be queried */
	for(j=0;j<b->bloom_bits;j++){
	    v = (v<<1) | get_bit(hash,offset+j);
	}
	if(nsrl_debug>1){printf("\t%08x (%d)  ",v,get_bloom_bit(b,v)); }
	if(get_bloom_bit(b,v)==0) return 0; /* if that bit wasn't set, return early */
    }
    /* All of the bits were set; hash must be in the bloom filter */
    if(nsrl_debug>1) putchar('\n');
    return 1;				/* must be present */
}

#ifndef O_BINARY
#define O_BINARY 0
#endif

#define xstr(s) str(s)
#define str(s) #s


nsrl_bloom *nsrl_bloom_open(const char *fname,int mode)
{
    char offset_buf[BLOOM_VECTOR_OFFSET];
    nsrl_bloom *b = calloc(sizeof(*b),1);
    char *line,*buf;
    int version = 0;
    int  prot = 0;

    switch(mode & O_ACCMODE){
    case O_RDONLY: prot = PROT_READ;if(nsrl_debug) printf("PROT_READ\n");break;
    case O_WRONLY: prot = PROT_WRITE;if(nsrl_debug) printf("PROT_WRITE\n");break;
    case O_RDWR:   prot = PROT_READ|PROT_WRITE;if(nsrl_debug) printf("PROT_READ|PROT_WRITE\n");break;
    }

    /* Open the file and get the parameters */

    b->fd = open(fname,mode|O_BINARY,0);
    if(b->fd<0){
	free(b);
	return 0;			/* could not open */
    }

    if(read(b->fd,offset_buf,sizeof(offset_buf))!=sizeof(offset_buf)){
	close(b->fd);
	free(b);
	errno = EINVAL;
	return 0;
    }
    buf = offset_buf;
    
    while((line = strsep(&buf,"\n"))){
	char *colon = index(line,':');
	if(colon){
	    *colon = 0;			/* terminate at the colon */
	    colon++;
	    if(strcmp(line,str(nsrl_bf_version))==0) version = atoi(colon);
	    if(strcmp(line,str(hash_bits))==0)       b->hash_bits = atoi(colon);
	    if(strcmp(line,str(bloom_bits))==0)      b->bloom_bits = atoi(colon);
	    if(strcmp(line,str(k))==0)  b->k = atoi(colon);
	    if(strcmp(line,str(added_items))==0)     b->added_items = atoi(colon);
	    if(strcmp(line,str(comment))==0)         b->comment = strdup(colon);
	}
    }
    if(version!=2){
	fprintf(stderr,"require nsrl bf vesion 2; got version %d\n",version);
	free(b);
	errno= EINVAL;
	return 0;
    }
    if(b->hash_bits==0 || b->bloom_bits==0){
	close(b->fd);
	if(b->comment) free(b->comment);
	free(b);
	errno= EINVAL;
	return 0;
    }
    b->vector_bytes  = 1 << (b->bloom_bits-3);
    b->vector = mmap(0,b->vector_bytes,
		     prot,
		     MAP_FILE|MAP_SHARED,
		     b->fd,BLOOM_VECTOR_OFFSET);
    if(b->vector == MAP_FAILED){
	close(b->fd);
	if(b->comment) free(b->comment);
	free(b);
	return 0;
    }
    b->memmapped = 1;
    //if(nsrl_debug) nsrl_histogram(b);
    return b;
}

/** nsrl_bloom_create:
 * @param fname - name of the filter to create; 0 if in-memory
 * @param hash_bits  - the number of bits in values to be hashed.
 * @param bloom_bits - log2(m)
 * @param k = k
 * @param comment - a comment to store in the bloom filter.
 */

nsrl_bloom *nsrl_bloom_create(const char *fname,int hash_bits, int bloom_bits,int k,
			      const char *comment)
{
    int vector_bytes = 1 << (bloom_bits - 3); /* Needed vector size */
    unsigned char *vector_buf = calloc(vector_bytes,1);
    nsrl_bloom *b = calloc(sizeof(*b),1);

    if(vector_buf==0) return 0;		/* not enough memory to allocate vector buf */
    b = calloc(sizeof(*b),1);
    if(b==0){				/* pretty weird */
	free(vector_buf);
	return 0;
    }

    if(hash_bits < bloom_bits * k) err(1,"bloom_bits * k > hash_bits (%d * %d > %d)",bloom_bits,k,hash_bits);
    if(vector_bytes < (1<<(bloom_bits-3))) err(1,"vector isn't big enough (%d * 8 < 2^(%d-1))",
					       vector_bytes,bloom_bits);

    b->hash_bits     =  hash_bits;
    b->bloom_bits    =  bloom_bits;
    b->k             = k;
    b->vector_bytes  =  vector_bytes;
    b->vector        =  vector_buf;
    b->comment       =  strdup(comment);

    if(fname==0) return b;		/* nameless bloom filter */

    /* Create the file with the paremters and empty bloom filter */
    b->fd = open(fname,O_CREAT|O_RDWR|O_EXCL|O_BINARY,0666); /* open */
    if(b->fd<0){
	free(vector_buf);
	free(b);
	return 0;		/* failed */
    }
    if(nsrl_bloom_info_update(b)){
	close(b->fd);
	unlink(fname);
	free(vector_buf);
	return 0;			/* some writing problem */
    }
    lseek(b->fd,BLOOM_VECTOR_OFFSET,SEEK_SET);
    if(write(b->fd,vector_buf,vector_bytes)!=vector_bytes){
	close(b->fd);
	unlink(fname);
	free(vector_buf);
	return 0;			/* malloc problem */
    }
    /* Now map the bloom filter into memory */
    free(vector_buf);
    b->vector = mmap(0,b->vector_bytes,PROT_READ|PROT_WRITE,MAP_FILE|MAP_SHARED,b->fd,BLOOM_VECTOR_OFFSET);
    return b;
}

int nsrl_bloom_free(nsrl_bloom *b)
{
    if(b->memmapped) munmap(b->vector,b->vector_bytes);
    else free(b->vector);		/* free the memory otherwise */
    if(b->comment) free(b->comment);
    memset(b,0,sizeof(*b));		/* clean object reuse */
    free(b);
    return 0;
}
