/*
 * fiwalk.cpp:
 * File and Inode Walk.
 *
 * This application uses SleuthKit to generate a report of all of the files
 * and orphaned inodes found in a disk image. It can optionally compute the MD5 of 
 * any objects, save those objects into a directory, or both.
 *
 * Algorithm:
 * 1 - Find all of the partitions on the disk. 
 * 2 - For each partition, walk the files.
 * 3 - For each file, print the requested information.
 * 4 - For each partition, walk the indoes
 * 5 - For each inode, print the requested information.
 */


#include "config.h"

#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/time.h>

#include <regex.h>
#include <ctype.h>
#include <fcntl.h>

#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

#ifdef HAVE_INTTYPES_H
#ifndef __STDC_FORMAT_MACROS
#define __STDC_FORMAT_MACROS
#endif
#include <inttypes.h>
#endif

#include <err.h>

#include <openssl/evp.h>
#include <openssl/md5.h>
#include <openssl/sha.h>
#include <tsk/libtsk.h>

#ifdef HAVE_AFFLIB
#include <afflib/afflib.h>
#endif

#include <algorithm>
#include <cstdlib>
#include <vector>
#include <string>
using namespace std;


/* Bring in our headers */
#include "arff.h"
#include "plugin.h"
#include "bloom.h"
#include "bloomset.h"

#include "threaded_hash.h"
#include "tagset.h"
#include "xml.h"


/* Structure for keeping track of file segments */
struct seg {
    int64_t start;			// byte offset from start of image
    int64_t len;				// number of bytes
};

/* Configurable options */
char *fiwalk_version = "0.4";
int  do_orphan=0;
int  do_allocated=0;
char *save_outdir = "";
int  opt_debug    = 0;
int  opt_md5      = 0;			// calculate the MD5 for every file?
int  opt_sha1     = 0;			// calculate the SHA1 for every file?
bool opt_save     = false;
bool opt_userguess = false;
int  opt_multithreaded = 1;		// use multiple threading
bool do_plugin = false;		// this one requires plugin processing
char *save_bloom512 = 0;
const char *config_file = 0;

bloomset mybloomset;

typedef vector<struct seg> seglist;	// vector of blocks
seglist segs;

vector<string> namelist;		// names of files that we want to find

/****************************************************************
 ** Current information.
 **
 ****************************************************************/
int      current_partition_num=0;
int64_t  current_partition_start=0;	// in bytes

/* Individual 'state' variables */
FILE   *plugin_outfile = 0;		// where to send for plugin processing
string  plugin_filename;

void usage()
{
    printf("usage: fiwalk [options] iso-name\n");
    printf("Default behavior: Just print the file system statistics and exit.\n");
    printf("options:\n");
    printf("    -F = report allocated files\n");
    printf("    -O = report orphan files  (default is \n");
    printf("    note: if neither -F nor -O is provided, -FO is assumed\n");
    printf("    -c config.txt = read config.txt for metadata extraction tools\n");
    printf("    -u       = Guess at usernames based on filenames, mode, or UIDs\n");
    printf("    -S     Single threaded; do not use multi-threaded.\n");
    
    printf("\n");
    printf("include/exclude parameters; may be repeated. \n");
    printf("NOTE: DO NOT RELY ON THIS SYNTAX; IT WILL CHANGE (soon)\n");

    printf("    -b+bloomfile.bloom = Only include files whose SHA1 is in bloomfile.bloom\n");
    printf("    -b-bloomfile.bloom = Only include files whose SHA1 is not in bloomfile.bloom\n");
    printf("    -n pattern  = only match files for which the filename matches\n");
    printf("                  the pattern.\n");
    printf("              example: -n .jpeg -n .jpg will find all JPEG files\n");
    printf("              Case is ignored. Will not match orphan files.\n");
    printf("    ");
    printf("\n");
    printf("\nFull content options (slow processing significantly):\n");
    printf("    -m = Also report MD5 for each file\n");
    printf("    -1 = Also report SHA1 for each file\n");
    printf("    -s <dir> = Save all recovered files to <dir>\n");
    printf("\n");
    printf("Output options:\n");
    printf("    -A = ARFF output to stdout\n");
    printf("    -X file = XML output to a given file \n");
    printf("    -B = output 512-byte bloom filter\n");
    printf("\n");
    printf("    -d = debug this program\n");
    printf("    -v = Enable SleuthKit verbose flag\n");
    printf("\n");
#ifdef HAVE_AFFLIB
    printf("AFFLIB Version %s\n",af_version());
#else
    printf("*** NO AFFLIB SUPPORT ***\n");
#endif    
    printf("\n");
    exit(1);
}

/****************************************************************
 ** Support routines 
 ****************************************************************/
static const char *cstr(const string &str){
    return str.c_str();
}

static string empty("");


/****************************************************************
 ** XML output
 ****************************************************************/

/****************************************************************
 ** Metadata Output
 ****************************************************************/

/* Specify the output mode */
enum  output_mode { ARFF, XML, TEXT };
enum  output_mode out = TEXT;		// default
class arff *a=0;			// ARFF generator
class xml  *x = 0;

/**
 * output a comment in the current file format
 */

void comment(const char *format,...)
{
    switch(out){
    case ARFF: break;
    case TEXT: fputs("# ",stdout);break;
    case XML:  fputs("<!-- ",x->file());break;
    }

    char buf[1024];
    va_list ap;
    va_start(ap, format);
    switch(out){
    case ARFF:
	vsnprintf(buf,sizeof(buf),format,ap);
	a->add_comment(buf);
	break;
    case TEXT:
	vprintf(format,ap);
	break;
    case XML:
	vfprintf(x->file(),format,ap);
	break;
    }
    va_end(ap);

    switch(out){
    case ARFF: break;
    case TEXT: putchar('\n');break;
    case XML:  fputs(" -->\n",x->file());break;
    }
}

void xmlout(const string &name,const string &value,const string &attribute)
{
    x->xmlout(xml::xmlstrip(name),xml::xmlescape(value),attribute);
}




/**
 * output a name/format/value for the current partition.
 * This information is simply printed as comments for ARFF files.
 *
 * @param name - the name of the thing being output
 * @param format - the format for the thing
 * @param ... - the value
 * This will output as a comment in the ARFF file
 */
void partition_info(const string &name,const string &value,const string &attribute)
{
    switch(out){
    case ARFF:
	a->add_comment(name + ": " + value);
	return;
    case TEXT:
	fputs(cstr(name + ": " + value + "\n"),stdout);
	return;
    case XML:
	xmlout(name,value,attribute);
	return;
    }
}

void partition_info(const char *name,const char *value)
{
    partition_info(string(name),string(value),empty);
}

void partition_info(const string &name,long i)
{
    char buf[1024];
    sprintf(buf,"%d",i);
    partition_info(name,buf,empty);
}

/****************************************************************/
void file_info(const string &name,const string &value,const string &attribute)
{
    switch(out){
    case ARFF: a->add_value(name,value); return;
    case TEXT: fputs(cstr(name + ": " + value + "\n"),stdout); return;
    case XML: xmlout(name,value,attribute); return;
    }
}

void file_info(const string name, long long v)
{
    char buf[1024];
    snprintf(buf,sizeof(buf),"%"PRIuDADDR,v);
    file_info(name,buf,empty);
}

void file_infot(const string name,time_t t)
{
    switch(out){
    case ARFF: a->add_valuet(name,t); return;
    case XML: file_info(name,t); return;
    default:
	file_info(name,t);
	char buf[64];
	strftime(buf,sizeof(buf),"%F %T",gmtime(&t));
	file_info(name+"_txt",buf,empty);
	return;
    }
}

void file_info(const char *name,const char *value)
{
    file_info(string(name),string(value),empty);
}


/****************************************************************/
string mytime()
{
    time_t t = time(0);
    char *buf = ctime(&t);
    buf[24] = 0;
    return string(buf);
}

/****************************************************************
 ** Content Output
 ****************************************************************/

class content_info {
public:
    FILE   *outfile;
    string   current_filename; // filename of what's currently being saved
    string   current_path;     // full path of file currently being processed
    string   save_path;	       // full path of where it is being saved
    threaded_hash h_md5;
    threaded_hash h_sha1;
    content_info():h_md5(EVP_md5,opt_multithreaded),
		   h_sha1(EVP_sha1,opt_multithreaded){
	outfile = 0;
	if(outfile){
	    fclose(outfile);
	    outfile = 0;
	}
    }
    void open_savefile(const char *filename);
    void add_bytes(const char *buf,size_t size){
	if(opt_md5)   h_md5.update((unsigned char *)buf,size);
	if(opt_sha1)  h_sha1.update((unsigned char *)buf,size);
	if(opt_save)  fwrite(buf,size,1,outfile);
    }
    void final(){
	if(opt_save){
	    fclose(outfile);
	    outfile = 0;
	}
    }
};

/** Open a file where a bytestream will be saved.
 *
 * @param filename - specifies the name of the file. If the file
 * exists, then start a counter between the end of the file and the
 * extension and keep incrementing...
 *
 */
void content_info::open_savefile(const char *filename)
{
    if(access(save_path.c_str(),F_OK)!=0){
	outfile = fopen(save_path.c_str(),"w");
	return;
    }

    char fmt[1024];
    const char *ext = strrchr(filename,'.');

    strcpy(fmt,filename);
    char *cc = strchr(fmt,'.');
    if(cc) *cc = '\000';
    strcat(fmt,".%d");
    if(ext) strcat(fmt,ext);

    for(int i=1;i<10000;i++){
	char fname[1024];
	snprintf(fname,sizeof(fname),fmt,i);
	if(access(fname,F_OK)!=0){
	    outfile = fopen(fname,"w");
	    return;
	}
    }
    // hm. Can't open file?
}


/****************************************************************
 *** Support Code.
 ****************************************************************/

/* FS_INFO defined in ~/sleuthkit-2.05/src/fstools/fs_tools.h
 * IMG_INFO defined in img_tools.h 
 * img_open - opens the image; calls do_dimage
 * do_dimage - analyizes the image
 *        mm_open - opens the volume; mm_part_walk - walk each volume;
 *                  calls mm_act for each
 * mm_act - analyizes each partition; calls do_vol for ones that we can analyze
 * do_vol - analyizes each volume
 *       fs_open - opens the file system; dent_walk 
 *               - walks the directory structure.
 *               - calls dent_act for each file
 * dent_act - prints file name; calls file_walk() to read each file; 
 *          - calls file_act for each sector
 * file_act - prints the block number
 */

/** 
 * file_act:
 *  -- obtain the address of each data unit
 *  -- optionally calculate the mac or save the data
 * 
 */
TSK_WALK_RET_ENUM
file_act(TSK_FS_INFO * fs, TSK_DADDR_T addr, char *buf, size_t size, TSK_FS_BLOCK_FLAG_ENUM flags, void *ptr)
{
    //printf("file_act: addr=%qd buf=%x size=%d flags=%x\n",addr,buf,size,flags);

    content_info *ci = (content_info *)ptr;

    if(opt_debug){
	printf("addr=%"PRIuDADDR" buf=%p size=%d\n",addr,buf,(int)size);
	if(opt_debug>1 && segs.size()==0){
	    fwrite(buf,size,1,stdout);
	    printf("\n");
	}
    }

    if(size==0)  return TSK_WALK_CONT;	    // can't do much with this...
    ci->add_bytes(buf,size);
    if(plugin_outfile) fwrite(buf,size,1,plugin_outfile);

    /* See if we can extend the last segment in the segment list,
     * or if this is the start of a new fragment.
     */
    int64_t byteaddr = current_partition_start + ((int64_t)addr)*fs->block_size;

    if(segs.size()>0){
	if(segs.back().start + segs.back().len == (int64_t)byteaddr){
	    segs.back().len += size;
	    return TSK_WALK_CONT;
	}
    }
    /* Need to add a new element to the list */
    struct seg newseg;
    newseg.start = byteaddr;
    newseg.len   = size;
    segs.push_back(newseg);
    return TSK_WALK_CONT;
}


/**
 * Try to guess the username from the path and, if found, add it.
 */
void add_userguess(class arff *a,const string &path)
{
    static bool first = true;
    static regex_t re;
    if(first){
	if(regcomp(&re,"Documents and Settings/([^/]+)/",REG_EXTENDED|REG_ICASE)) err(1,"regcomp");
	first = 0;
    }
    regmatch_t pmatch[4];
    memset(pmatch,0,sizeof(pmatch));
    if(regexec(&re,path.c_str(),2,pmatch,0)) return;
    string username = path.substr(pmatch[1].rm_so,pmatch[1].rm_eo-pmatch[1].rm_so);
    a->add_value("userguess",username);
}


/*
 * Common accounting routine for both inode searching and dent searching.
 * Called for each file in the file system, whether it is a named file or an orphan file. 
 * (ptr==NULL for orphans).
 *
 * Walk each file to print the addresses
 */
static TSK_WALK_RET_ENUM
inode_act(TSK_FS_INFO * fs, TSK_FS_INODE * fsi, void *ptr)
{
    content_info ci;

    /** First, decide if we should return */
    /* If the name has corresponding metadata, then walk it */
    if(!fsi) return TSK_WALK_CONT;	// no file system info??

    if ((fsi->mode & TSK_FS_INODE_MODE_FMT)!=TSK_FS_INODE_MODE_REG){
	/* Only do regular files */
	return TSK_WALK_CONT;
    }

    /* Recover the filename from the fs_dent, if it is provided */
    TSK_FS_DENT *fs_dent = (TSK_FS_DENT *)ptr;
    if (fs_dent) {
	ci.current_filename = fs_dent->name;
	ci.current_path = string(fs_dent->path) + ci.current_filename;
	if(opt_save) ci.save_path = string(save_outdir) + "/" + ci.current_filename;
    }


    /* If we are filtering and we have a filename, see if we want this file. */
    if(namelist.size()>0 && ci.current_filename.size()>0){
	bool file_wanted = false;
	for(vector<string>::const_iterator i = namelist.begin(); i!=namelist.end(); i++){
	    if(strcasestr(ci.current_filename.c_str(),i->c_str())){
		file_wanted = true;
	    }
	}
	if(!file_wanted) return TSK_WALK_CONT;
    }

    /* Looks like we are processing */
    switch(out){
    case ARFF: a->new_row();break;
    case XML:  x->push("fileobject"); break;
    case TEXT: break;
    }
	    
    /* See if this one requires plugin processing */
    do_plugin = plugin_match(ci.current_filename.c_str());
    if(do_plugin){
	plugin_filename = ci.current_filename;
	/* Remove special characters */
	for(int i=0;i<plugin_filename.size();i++){
	    if(isalnum(plugin_filename[i])==0) plugin_filename[i]='_';
	}
	plugin_filename = "/tmp/" + plugin_filename;
	plugin_outfile = fopen(plugin_filename.c_str(),"w");
	if(!plugin_outfile) errx(1,plugin_filename.c_str());
    }

    // do we want fully content?
    int full_content = (opt_md5) || (opt_sha1) || (opt_save) || (do_plugin)  || (mybloomset.size()>0); 
    segs.clear();				// segments of where the file is

    int myflags = TSK_FS_FILE_FLAG_NOID;
    myflags |= TSK_FS_FILE_FLAG_RECOVER; // use special data recovery for deleted files

    if(opt_save)	ci.open_savefile(ci.save_path.c_str());

    if(!full_content) {
	myflags |= TSK_FS_FILE_FLAG_AONLY;	/* Just get the metadata */
    }
	
    /* Now get the information on the individual
     * sectors that make up the file.
     */
    if (fs->file_walk(fs, fsi, 0, 0, (TSK_FS_FILE_FLAG_ENUM)myflags, file_act, (void *)&ci)) {
	tsk_error_reset();
    }
	
    /* See if the bloomset says to get rid of the file */
    if(mybloomset.size()>0){
	u_char sha1buf[20];
	ci.h_sha1.final(sha1buf,sizeof(sha1buf));
	if(mybloomset.check_exclude(sha1buf)){
	    unlink(plugin_filename.c_str());
	    return TSK_WALK_CONT;
	}
    }

    /* Send through to the plugin if we were doing that. */
    if(do_plugin){
	fclose(plugin_outfile);
	plugin_outfile = 0;
	plugin_process(plugin_filename.c_str(),a);
	unlink(plugin_filename.c_str());
    }

    /* Finally output the informaton */
    if(ci.current_path.size()>0) file_info("Filename",ci.current_path,empty);
    if(ptr==NULL) file_info("Orphan","YES");
    file_info("Filesize",fsi->size);
    if(fsi->flags & TSK_FS_INODE_FLAG_ALLOC)   file_info("ALLOC",1);
    if(fsi->flags & TSK_FS_INODE_FLAG_UNALLOC) file_info("UNALLOC",1);
    if(fsi->flags & TSK_FS_INODE_FLAG_USED)    file_info("USED",1);
    if(fsi->flags & TSK_FS_INODE_FLAG_UNUSED)  file_info("UNUSED",1);
    if(fsi->flags & TSK_FS_INODE_FLAG_ORPHAN)  file_info("ORPHAN",1);
    if(fsi->mtime){
	file_infot("mtime",fsi->mtime);
    }
    if(fsi->ctime){
	file_infot("ctime",fsi->ctime);
    }
    if(fsi->atime){
	file_infot("atime",fsi->atime);
    }
    if(segs.size()>0){
	string runz = "";
	for(seglist::const_iterator i = segs.begin();i!=segs.end();i++){
	    char buf[1024];
	    sprintf(buf,"%qd:%qd",i->start,i->len);
	    if(runz.size()>0) runz += "|";
	    runz += buf;
	}
	file_info("Byte runs",runz,empty);
    }
    file_info("Fragments",segs.size());
    if(opt_md5){
	u_char md5buf[16];
	char buf[64];
	ci.h_md5.final(md5buf,sizeof(md5buf));
	nsrl_hexbuf(buf,sizeof(buf),md5buf,16,0);
	file_info("md5",buf);
    }
    if(opt_sha1){
	u_char sha1buf[20];
	char buf[64];
	ci.h_sha1.final(sha1buf,sizeof(sha1buf));
	nsrl_hexbuf(buf,sizeof(buf),sha1buf,20,0);
	file_info("sha1",buf);
    }
    
    file_info("partition",current_partition_num);
    if(fs->block_size>0){
	if(segs.size()>=1) file_info("frag1startsector",(int)(segs[0].start / fs->block_size));
	if(segs.size()>=2) file_info("frag2startsector",(int)(segs[1].start / fs->block_size));
    }
    if(ci.current_path.size()>0){
	if(opt_userguess) add_userguess(a,ci.current_path);
    }


    /* END of file processing */
    switch(out){
    case ARFF: break;
    case XML:  x->pop("fileobject");break;
    case TEXT: fputs("\n",stdout);break;
    }
    fflush(stdout);
    return TSK_WALK_CONT;
}

/*
 * dent_acct:
 * Called for each file in the file system
 * Callback action for the file system dent_walk
 *
 * Walk each file to print the addresses
 */
TSK_WALK_RET_ENUM
dent_act(TSK_FS_INFO * fs, TSK_FS_DENT * fs_dent, void *ptr)
{
    if (fs_dent->fsi) {
	return inode_act(fs,fs_dent->fsi,(void *)fs_dent);
    }
    return TSK_WALK_CONT;
}


/* 
 * Analyze the volume starting at byte offset 'start' in the image
 * file pointed to by img. 
 *
 * Return 0 on success and 1 on error 
 */
int do_vol(TSK_IMG_INFO *img, off_t start)
{
    TSK_FS_INFO *fs;
    int ret = 0;

    if (tsk_verbose) fprintf(stderr, "do_vol:  Start: %" PRIuDADDR "\n", start);

    if(out==XML) x->printf("<volume startsector='%"PRIuDADDR"'>\n");

    /* Try it as a file system */
    if ((fs = tsk_fs_open(img, start, NULL)) == NULL) {
	partition_info("TSK Error",tsk_error_get());
	if(out==XML) x->pop("volume");
	/* @@@ We could do some file system carving at this point */
	return 1;
    }


    current_partition_num++;
    current_partition_start = fs->offset;
    if(out==TEXT || out==XML){
	partition_info("Partition Offset",fs->offset);
	partition_info("block size",fs->block_size);
	partition_info("ftype",fs->ftype);
	partition_info("ftype_str",tsk_fs_get_type(fs->ftype));
	partition_info("block_count",fs->block_count);
	partition_info("first_block",fs->first_block);
	partition_info("last_block",fs->last_block);
	if(out==TEXT) putchar('\n');
	fflush(stdout);
    }


    /*
     * If we didn't want information on the files, we could just
     * return here with
     *
     * fs->close(fs); return 0;
     */
    
    /* Walk the files, starting at the root directory */
    if(do_allocated){
	time_t t = time(0);
	comment("Starting dent_walk");
	int dent_walk_flags = (TSK_FS_DENT_FLAG_ALLOC | TSK_FS_DENT_FLAG_UNALLOC |
			       TSK_FS_DENT_FLAG_RECURSE);
	if (fs->dent_walk(fs, fs->root_inum, (TSK_FS_DENT_FLAG_ENUM)dent_walk_flags, dent_act, &ret)) {
	    comment("TSK Error dent_walk: ",tsk_error_get());
	    ret = 1;
	}
    }

    if(do_orphan){
	time_t t = time(0);
	comment("Performing orphan inode_walk from %qd to %qd ",
		(long long)fs->first_inum,(long long)fs->last_inum);

	/* Don't need other flags; it will fill in the rest */
	TSK_FS_INODE_FLAG_ENUM inode_walk_flags = TSK_FS_INODE_FLAG_ORPHAN; 
	if (fs->inode_walk(fs,fs->first_inum,fs->last_inum,inode_walk_flags, inode_act, NULL)){
	    comment("TSK Error inode_walk",tsk_error_get());
	    ret = 1;
	}
    }

    fs->close(fs);
    if(out==XML) x->printf("</volume>\n");
    return ret;
}

/*
 * Callback for the partition table walk.
 *
 * Analyze the file system for each partition.
 *
 */
TSK_WALK_RET_ENUM mm_act(TSK_MM_INFO * mm, TSK_PNUM_T pnum, TSK_MM_PART * part, int flag, void *ptr)
{
    // Currently, skip the "descriptive" entries
    if (part->type == TSK_MM_PART_TYPE_DESC){
	return TSK_WALK_CONT;
    }

    if (do_vol(mm->img_info, part->start * mm->block_size)) {
	return TSK_WALK_ERROR;
    }
    return TSK_WALK_CONT;
}



/* do_dimage:
 * Process the partition tables in the disk image pointed to by img.
 * File system analysis will be performed on each partition.
 *   mm_open() - opens partition
 *   mm_part_walk() - walks partition
 * If we can't open as a disk image, try to open as a raw file system
 *
 * returns 1 on error and 0 on success
 */
int do_dimage(TSK_IMG_INFO * img, char *desc)
{
    TSK_MM_INFO *mm;

    if ((mm = tsk_mm_open(img, 0, NULL)) == NULL) {
	if (tsk_verbose) fprintf(stderr, "Error determining volume system -- trying file systems\n");

	/* Check if we have a partition image */
 	tsk_error_reset();
	if (do_vol(img, 0)) {
	    return 1;
	}
    }
    else {
	if (tsk_verbose) fprintf(stderr, "Volume system open, examining each\n");
	/* Walk the volumes and update the volume table */
	if (mm->part_walk(mm, mm->first_part, mm->last_part, 0, mm_act, 0)) {
	    return 1;
	}
    }
    return 0;
}




void process_bloomfile(const char *fname)
{
    if(fname[0]!='+' && fname[0]!='-'){
	fprintf(stderr,"usage: -b[+/-]bloomfile.bloom\n");
	usage();
    }
    bloomset_element::Action action = (fname[0]=='+' ? bloomset_element::INCLUDE : bloomset_element::EXCLUDE);
    bloomset_element *be = new bloomset_element(action, fname+1);
    mybloomset.push_back(be);
}

int main(int argc, char *argv[])
{
    TSK_IMG_INFO *img;
    int ch;
    extern int optind;

    int t0 = time(0);

    while ((ch = getopt(argc, argv, "mdvOF1s:AX:n:c:ub:B:S")) > 0) {
	switch (ch) {
	case '1': opt_sha1++;break;
	case 'm': opt_md5++; break;
	case 'd': opt_debug++; break;
	case 'v': tsk_verbose++; break; 			// sleuthkit option	  
	case 'F': do_orphan = 1; break;
	case 'O': do_allocated=1; break;
	case 's': save_outdir = optarg; opt_save = true; break;
	case 'A': out=ARFF ; break;
	case 'X':
	    out=XML;
	    x = new xml();
	    x->setoutfile(optarg);
	    break;

	case 'n': namelist.push_back(optarg);break;
	case 'c': config_file = optarg; break;
	case 'u': opt_userguess = true; break;
	case 'B': save_bloom512 = optarg; break;
	case 'b': process_bloomfile(optarg);break;
	case 'S': opt_multithreaded = 0;break;
	case 'h':
	case '?':
	default:
	    fprintf(stderr, "Invalid argument: %s\n", argv[optind]);
	    usage();
	}
    }
    if(config_file) config_read(config_file);    /* Read the configuration file */

    /* If neither -O and -F are specified, do both! */
    if(do_orphan==0 && do_allocated==0){
	do_orphan = 1;
	do_allocated = 1;
    }

    if (optind >= argc) usage();
    argc -= optind;
    argv += optind;

    const char *filename = argv[0];

    if(strstr(filename,".aff")){
#ifndef HAVE_AFFLIB
	fprintf(stderr,"ERROR: fiwalk was compiled without AFF support.\n");
	exit(0);
#else
	if(tsk_img_get_type(TSK_IMG_INFO_TYPE_AFF_AFF)==0){
	    fprintf(stderr,"ERROR: fiwalk was compiled with AFF support but the TSK library is not.\n");
	    exit(0);
	}
#endif
    }
    

    /* If we are outputing ARFF, create the ARFF object and set the file types for a few */
    switch(out){
    case ARFF:
	a = new arff("fiwalk");		// the ARFF output object
	a->add_attribute("id",arff::NUMERIC);
	a->add_attribute("partition",arff::NUMERIC);
	a->add_attribute("filesize",arff::NUMERIC);
	a->add_attribute("mtime",arff::DATE);
	a->add_attribute("ctime",arff::DATE);
	a->add_attribute("atime",arff::DATE);
	a->add_attribute("fragments",arff::NUMERIC);
	a->add_attribute("frag1startsector",arff::NUMERIC);
	a->add_attribute("frag2startsector",arff::NUMERIC);
	a->add_attribute("filename",arff::STRING);
	if(opt_md5) a->add_attribute("md5",arff::STRING);
	if(opt_sha1) a->add_attribute("sha1",arff::STRING);
	if(opt_userguess) a->add_attribute("userguess",arff::STRING);
	break;
    case TEXT:
	break;
    case XML:
	x->push("fiwalk");
	break;
    }

    /* Check that we have a valid file format */
    partition_info("fiwalk version",fiwalk_version);
    partition_info("Start time",mytime(),empty);


    if (opt_debug) printf("calling tsk_img_open(%s)\n",filename);
    if ((img = tsk_img_open(NULL, argc, (const char **)argv)) == NULL) {
	comment("TSK Error (img_open) %s",tsk_error_get());
	comment("=EOF= (img_open)");
	exit(1);
    }
    comment("image filename: %s",filename);

#ifdef HAVE_AFFLIB
    AFFILE *af = af_open(filename,0,O_RDONLY);
    if(!af) af_err(1,"af_open(%s)",filename);
    
    comment("AFF version: %s",af_version());
    u_char gid[1024];
    char buf[1024];
    size_t gidlen = sizeof(gid);
    if(af_get_seg(af,AF_IMAGE_GID,0,gid,&gidlen)==0){
	comment("AF_IMAGE_GID",af_hexbuf(buf,sizeof(buf),gid,gidlen,AF_HEXBUF_NO_SPACES));
    }


#endif    

    partition_info("Imagefile",filename);
    if (opt_debug) printf("calling do_dimage()\n");
    if (do_dimage(img, "my boring test comment")) {
	comment("TSK Error (do_dimage) %s",tsk_error_get());
    }
    img->close(img);

#ifdef HAVE_GETRUSAGE
    /* Print usage information */
    struct rusage ru;
    memset(&ru,0,sizeof(ru));
    int t1 = time(0);
    comment("clock: %d",t1-t0);
    if(getrusage(RUSAGE_SELF,&ru)==0){
	partition_info("usage user seconds",ru.ru_utime.tv_sec);
	partition_info("usage system seconds",ru.ru_stime.tv_sec);
	partition_info("usage maxrss",ru.ru_maxrss);
	partition_info("usage reclaims",ru.ru_minflt);
	partition_info("usage faults",ru.ru_majflt);
	partition_info("usage swaps",ru.ru_nswap);
	partition_info("usage inputs",ru.ru_inblock);
	partition_info("usage outputs",ru.ru_oublock);
	partition_info("stop time",cstr(mytime()));
    }
#endif
    switch(out){
    case ARFF:
	a->output(stdout);
	break;
    case TEXT:
	comment("=EOF=");
	break;
    case XML:
	x->pop("fiwalk");
	x->write();
    }
    exit(0);
}


